#include "hft_interface.h"

void hft_interface()
{
  GtkWidget 
    *window,
    *box,
    *scale,
    *label,
    *spin_button,
    *button;
  
  /* GtkAdjustment incompatible pointer type? */
  GtkObject 
    *displacement_adj, 
    *row_adj,
    *col_adj,
    *update_delay_adj;

  static gfloat F_row_sample;
  static gfloat F_col_sample;

  /* the max between height and width used for setting the scale */
  gfloat sample_max;

  F_row_sample = (gfloat) hft_current.row_sample_rate;
  F_col_sample = (gfloat) hft_current.col_sample_rate;

  if (hft_drawable->height > hft_drawable->width) 
    sample_max = (float) hft_drawable->height/4;
  else
    sample_max = (float) hft_drawable->width/4;

  /* end some random sample rate code */

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (window), "displacement thingy");
  gtk_signal_connect (GTK_OBJECT (window), "destroy",
                      GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
  
  box = gtk_vbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(window), box);
  gtk_widget_show(box);

  /* scale */
  displacement_adj = gtk_adjustment_new 
    (hft_current.displacement_scale, 0, 3, .01, .1, 1);
  scale = gtk_hscale_new(GTK_ADJUSTMENT(displacement_adj));
  gtk_scale_set_digits(GTK_SCALE(scale),4);
  gtk_box_pack_start (GTK_BOX(box), scale, FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (displacement_adj), "value_changed",
		      GTK_SIGNAL_FUNC(scale_callback), 
		      &hft_current.displacement_scale);  
  gtk_widget_show(scale);
  
  label = gtk_label_new("vertical sample rate");
  gtk_box_pack_start (GTK_BOX (box), label, TRUE,TRUE, 0);
  gtk_widget_show (label);  
  row_adj = gtk_adjustment_new 
    (F_row_sample, 1, sample_max+1, 1, 1, 1);
  scale = gtk_hscale_new(GTK_ADJUSTMENT(row_adj));
  gtk_scale_set_digits(GTK_SCALE(scale),0);
  gtk_box_pack_start (GTK_BOX(box), scale, TRUE, TRUE, 1);
  gtk_signal_connect (GTK_OBJECT (row_adj), "value_changed",
		      GTK_SIGNAL_FUNC(row_sample_callback), 
		      &F_row_sample); 
  gtk_widget_show(scale);

  label = gtk_label_new("horizontal sample rate");
  gtk_box_pack_start (GTK_BOX (box), label, TRUE,TRUE, 0);
  gtk_widget_show (label);  
  col_adj = gtk_adjustment_new 
    (F_col_sample, 1, sample_max+1, 1, 1, 1);
  scale = gtk_hscale_new(GTK_ADJUSTMENT(col_adj));
  gtk_scale_set_digits(GTK_SCALE(scale),0);
  gtk_box_pack_start (GTK_BOX(box), scale, TRUE, TRUE, 1);
  gtk_signal_connect (GTK_OBJECT (col_adj), "value_changed",
		      GTK_SIGNAL_FUNC(col_sample_callback), 
		      &F_col_sample); 
  gtk_widget_show(scale);

  label = gtk_label_new("auto update delay");
  gtk_box_pack_start (GTK_BOX (box), label, TRUE,TRUE, 0);
  gtk_widget_show (label);  
  update_delay_adj = gtk_adjustment_new 
    (hft_current.update_interval, 400, 2001, 100, 100, 100);
  spin_button = gtk_spin_button_new (GTK_ADJUSTMENT(update_delay_adj), 0, 0);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_button), 
			    hft_current.update_interval);
  gtk_box_pack_start (GTK_BOX(box), spin_button, TRUE, TRUE, 1);
  gtk_signal_connect (GTK_OBJECT (update_delay_adj), "value_changed",
		      GTK_SIGNAL_FUNC(update_delay_callback), 
		      &hft_current.update_interval); 
  gtk_widget_show(spin_button);

  make_display_mode_menu(box);

  make_displacement_mode_menu(box);

  button = gtk_button_new_with_label("close");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
  gtk_box_pack_start(GTK_BOX(box),button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label("update");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      GTK_SIGNAL_FUNC (update_callback), NULL);
  gtk_box_pack_start(GTK_BOX(box),button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_button_new_with_label("export");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      GTK_SIGNAL_FUNC (hft_export_dialog), NULL);
  gtk_box_pack_start(GTK_BOX(box),button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  button = gtk_toggle_button_new_with_label("auto update");
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      GTK_SIGNAL_FUNC (auto_update_callback), NULL);
  gtk_box_pack_start(GTK_BOX(box),button, TRUE, TRUE, 0);
  gtk_widget_show(button);  
  
  gtk_widget_show(window);
  
}

static void make_display_mode_menu(GtkWidget *box)
{
  GtkWidget 
    *hbox,
    *label,
    *option_menu,
    *menu,
    *menu_item;  
  hbox = gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(box), hbox);
  gtk_widget_show(hbox);
  
  label = gtk_label_new ("color");
  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE,TRUE, 0);
  gtk_widget_show (label);
  
  option_menu = gtk_option_menu_new();
  menu = gtk_menu_new();

  menu_item = gtk_menu_item_new_with_label("image");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_image_point_callback), 
			     NULL); 
  
  menu_item = gtk_menu_item_new_with_label("foreground");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_foreground_point_callback), 
			     NULL); 

  menu_item = gtk_menu_item_new_with_label("gradient");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_gradient_point_callback), 
			     NULL); 

  menu_item = gtk_menu_item_new_with_label("image_line");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_image_line_callback), 
			     NULL); 
  /*
  menu_item = gtk_menu_item_new_with_label("image_fill");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_image_fill_callback), 
			     NULL); 
  */
  menu_item = gtk_menu_item_new_with_label("foreground line");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_foreground_line_callback), 
			     NULL); 

  menu_item = gtk_menu_item_new_with_label("gradient line");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_gradient_line_callback), 
			     NULL); 
  /*
  menu_item = gtk_menu_item_new_with_label("gradient fill");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (display_mode_gradient_fill_callback), 
			     NULL); 
  */
  gtk_box_pack_start (GTK_BOX(hbox), option_menu, TRUE, TRUE, 1);  
  gtk_widget_show( option_menu );
  gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
}

static void make_displacement_mode_menu(GtkWidget *box)
{
  GtkWidget 
    *hbox,
    *label,
    *option_menu,
    *menu,
    *menu_item;  

  hbox = gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(box), hbox);
  gtk_widget_show(hbox);
  
  label = gtk_label_new ("displacement mode");
  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE,TRUE, 0);
  gtk_widget_show (label);
  
  option_menu = gtk_option_menu_new();
  menu = gtk_menu_new();

  menu_item = gtk_menu_item_new_with_label("value");
  gtk_menu_append( GTK_MENU(menu), menu_item);
  gtk_widget_show( menu_item); 
  gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
			     GTK_SIGNAL_FUNC
			     (displacement_mode_value_callback), 
			     NULL); 
  
  if (hft_drawable->bpp == 2 || hft_drawable->bpp == 4)
    {
      menu_item = gtk_menu_item_new_with_label("alpha");
      gtk_menu_append( GTK_MENU(menu), menu_item);
      gtk_widget_show( menu_item); 
      gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
				 GTK_SIGNAL_FUNC
				 (displacement_mode_alpha_callback), 
				 NULL); 
    }

  if (hft_drawable->bpp > 2)
    {
      menu_item = gtk_menu_item_new_with_label("red");
      gtk_menu_append( GTK_MENU(menu), menu_item);
      gtk_widget_show( menu_item); 
      gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
				 GTK_SIGNAL_FUNC
				 (displacement_mode_red_callback), 
				 NULL); 
      
      menu_item = gtk_menu_item_new_with_label("green");
      gtk_menu_append( GTK_MENU(menu), menu_item);
      gtk_widget_show( menu_item); 
      gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
				 GTK_SIGNAL_FUNC
				 (displacement_mode_green_callback), 
				 NULL); 
      
      menu_item = gtk_menu_item_new_with_label("blue");
      gtk_menu_append( GTK_MENU(menu), menu_item);
      gtk_widget_show( menu_item); 
      gtk_signal_connect_object( GTK_OBJECT(menu_item), "activate",
				 GTK_SIGNAL_FUNC
				 (displacement_mode_blue_callback), 
				 NULL); 
    }

  gtk_box_pack_start (GTK_BOX(hbox), option_menu, TRUE, TRUE, 1);  
  gtk_widget_show( option_menu );
  gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
}
 

static void hft_export_dialog()
{
  GtkWidget *file_selection;

  file_selection = gtk_file_selection_new("export");
  
  gtk_signal_connect (GTK_OBJECT (file_selection), "destroy",
		      (GtkSignalFunc) gtk_widget_destroy,
		      GTK_OBJECT (file_selection)) ;
  
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_selection) 
				  ->ok_button), "clicked", 
		      (GtkSignalFunc) hft_export, file_selection);
  
  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION 
					 (file_selection)->cancel_button),
			     "clicked", (GtkSignalFunc) gtk_widget_destroy,
			     GTK_OBJECT (file_selection));
  
  gtk_file_selection_set_filename (GTK_FILE_SELECTION(file_selection), 
				   "penguin.wrl");
  
  gtk_widget_show(file_selection);
}

void set_auto_update(gboolean this_request)
{
  if (hft_current.update_set == TRUE)
    {
      if (this_request == TRUE)
	{
	  if(hft_current.update_count != 0) 
	    hft_current.update_count--;
	  if(hft_current.update_count == 0) 
	    {
	      hft_current.update_tag = gtk_timeout_add 
		(hft_current.update_interval, 
		 (GtkFunction)GTK_SIGNAL_FUNC(update_callback) , NULL);
#ifdef DEBUG
	      g_print("update interval %d\n",hft_current.update_interval);
#endif 
	    }
#ifdef DEBUG
	  g_print("update on count %d\n",hft_current.update_count); 
#endif
	}
      else 
	{
	  if (hft_current.update_count == 0)
	    {
	      gtk_timeout_remove(hft_current.update_tag);	
	    }
	  hft_current.update_count++;
#ifdef DEBUG
	  g_print("update off count %d\n",hft_current.update_count);
#endif	  
	}
    }
}

void auto_update_callback(GtkWidget *widget, gpointer data)
{
  if (GTK_TOGGLE_BUTTON (widget)->active) 
    {
      hft_current.update_set = TRUE;
      hft_current.update_count = 0;
      set_auto_update(TRUE);
    } 
  else 
    {
      hft_current.update_set = FALSE;
      gtk_timeout_remove(hft_current.update_tag);
    }
}


static void update_callback()
{
#ifdef DEBUG
  g_print("update\n");
#endif
  set_auto_update(FALSE); 
  hft_make_field();
  if (hft_current.display_mode==GRADIENT_POINT) 
    hft_get_gradient();
  hft_draw();
  set_auto_update(TRUE);
}

static void update_delay_callback(GtkAdjustment *adjustment, gint *val)
{
  set_auto_update(FALSE); 
  *val = adjustment->value;
  set_auto_update(TRUE);
#ifdef DEBUG
  g_print("delay now %d\n", hft_current.update_interval);
#endif
}

/* display modes */

static void display_mode_foreground_point_callback()
{
  hft_current.display_mode = FOREGROUND_POINT;
  hft_draw();
}

static void display_mode_gradient_point_callback()
{
  hft_current.display_mode = GRADIENT_POINT;
  hft_get_gradient();
  hft_draw();
}

static void display_mode_image_point_callback()
{
  hft_current.display_mode = IMAGE_POINT;
  hft_draw();
}

static void display_mode_image_line_callback()
{
  hft_current.display_mode = IMAGE_LINE;

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));

  hft_draw();
}

static void display_mode_foreground_line_callback()
{
  hft_current.display_mode = FOREGROUND_LINE;

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));

  hft_draw();
}

static void display_mode_gradient_line_callback()
{
  hft_current.display_mode = GRADIENT_LINE;

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));

  hft_get_gradient();
  hft_draw();
}

static void display_mode_image_fill_callback()
{
  hft_current.display_mode = IMAGE_FILL;

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));

  hft_get_gradient();
  hft_draw();
}

static void display_mode_gradient_fill_callback()
{
  hft_current.display_mode = GRADIENT_FILL;

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));

  hft_get_gradient();
  hft_draw();
}

/* displacement modes */

static void displacement_mode_value_callback()
{
  hft_current.displacement_mode = VALUE;
  hft_make_field();
  hft_draw();
}

static void displacement_mode_alpha_callback()
{
  hft_current.displacement_mode = ALPHA;
  hft_make_field();
  hft_draw();
}

static void displacement_mode_red_callback()
{
  hft_current.displacement_mode = RED;
  hft_make_field();
  hft_draw();
}

static void displacement_mode_green_callback()
{
  hft_current.displacement_mode = GREEN;
  hft_make_field();
  hft_draw();
}

static void displacement_mode_blue_callback()
{
  hft_current.displacement_mode = BLUE;
  hft_make_field();
  hft_draw();
}


static void scale_callback(GtkAdjustment *adjustment, gfloat *val)
{
  *val = adjustment->value;
  hft_draw();
}

static void row_sample_callback(GtkAdjustment *adjustment, gfloat *val)
{
  *val = adjustment->value;
  if (*val<1) hft_current.row_sample_rate = 1;
  else hft_current.row_sample_rate = (gint) *val;
  hft_make_field();
  hft_draw();
}

static void col_sample_callback(GtkAdjustment *adjustment, gfloat *val)
{
  *val = adjustment->value;
  if (*val<1) hft_current.col_sample_rate = 1;
  else hft_current.col_sample_rate = (gint) *val;
  hft_make_field();
  hft_draw();
}




