#include"hft_view.h"

void hft_view()
{
  GtkWidget *window;
  int glarea_attribute_list[] = { GDK_GL_RGBA,
				  GDK_GL_DOUBLEBUFFER,
				  GDK_GL_NONE };

  window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
  if (window == NULL)
    {
      g_print("view window never made it\n");
      gtk_main_quit();
    }
  gtk_window_set_title(GTK_WINDOW(window), "height field toy");
  gtk_container_border_width(GTK_CONTAINER(window), 0);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);  
  gtk_quit_add_destroy(1, GTK_OBJECT(window));

  hft_glarea = GTK_WIDGET(gtk_gl_area_new(glarea_attribute_list));

 if (hft_glarea == NULL)
    {
      g_print("glarea never made it\n");
      gtk_main_quit();
    }

  gtk_widget_set_events(GTK_WIDGET(hft_glarea),
                        GDK_EXPOSURE_MASK| 
			GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|
			GDK_POINTER_MOTION_MASK|
			GDK_POINTER_MOTION_HINT_MASK);
   gtk_signal_connect(GTK_OBJECT(hft_glarea), "expose_event",
                     GTK_SIGNAL_FUNC(expose_callback), NULL);
  gtk_signal_connect(GTK_OBJECT(hft_glarea), "configure_event",
                     GTK_SIGNAL_FUNC(reshape), NULL);
  gtk_signal_connect (GTK_OBJECT(hft_glarea), "motion_notify_event",
		      GTK_SIGNAL_FUNC(motion_notify), NULL);
  gtk_signal_connect (GTK_OBJECT(hft_glarea), "button_press_event",
		      GTK_SIGNAL_FUNC(button_press), NULL);
 gtk_signal_connect (GTK_OBJECT(hft_glarea), "button_release_event",
		      GTK_SIGNAL_FUNC(button_release), NULL);
  gtk_signal_connect(GTK_OBJECT(hft_glarea), "realize", 
                     GTK_SIGNAL_FUNC(hft_view_init), NULL);

  gtk_widget_set_usize(GTK_WIDGET(hft_glarea), 150,150);
  gtk_container_add(GTK_CONTAINER(window),GTK_WIDGET(hft_glarea));

  gtk_widget_show(GTK_WIDGET(hft_glarea));
  gtk_widget_show(GTK_WIDGET(window));

  hft_draw();
}

static void hft_view_init()
{  
  hft_current.row_sample_rate = hft_drawable->height / MESH_DENSITY;
  if (hft_current.row_sample_rate < 1) hft_current.row_sample_rate = 1;
     
  hft_current.col_sample_rate = hft_drawable->width / MESH_DENSITY;
  if (hft_current.col_sample_rate < 1) hft_current.col_sample_rate = 1;
  
  eye[X] = 0; 
  eye[Y] = 0; 
  eye[Z] = hft_drawable->height * 2.5;
  
  hft_make_field();
  hft_get_gradient();
  hft_gl_init();
}

static void hft_gl_init()
{
  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    {
      glViewport(0,0, hft_glarea->allocation.width, 
		 hft_glarea->allocation.height);
      glMatrixMode(GL_PROJECTION);
      gluPerspective(
		     hft_current.zoom,   
		     1.0,   
		     0.3,    
		     hft_drawable->width * 2);  
      gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));
    } 
}


void hft_make_field()
{
  GPixelRgn pix_r;
  guchar *pixel_buffer;
  
  hft_current.rowspan=hft_drawable->height / 
    hft_current.row_sample_rate;
  hft_current.colspan=hft_drawable->width / 
    hft_current.col_sample_rate;     
  
  if(hft_current.field != NULL)
    {
      g_free(hft_current.field);
    }

  hft_current.field = (char *) g_malloc (sizeof(char) 
					 * hft_current.rowspan
					 * hft_current.colspan);
  if (hft_current.field == NULL) 
    {
      g_print("height field toys could not allocate field\n");
      gtk_main_quit();
    }      

  
  gimp_pixel_rgn_init(&pix_r,hft_drawable,0,0,
		      hft_drawable->width,
		      hft_drawable->height,
		      FALSE,FALSE);
  
  pixel_buffer=(guchar *) g_malloc 
    (hft_drawable->height * hft_drawable->width * hft_drawable->bpp);
  if (pixel_buffer == NULL)
    {
      g_print("could not allocate buffer for drawable\n");
      gtk_main_quit();
    }

  gimp_pixel_rgn_get_rect(&pix_r,pixel_buffer,0,0,
			  hft_drawable->width,
			  hft_drawable->height);

  switch (hft_current.displacement_mode)
    {
    case ALPHA: hft_make_field_alpha(pixel_buffer); break;
    case VALUE: hft_make_field_value(pixel_buffer); break;
    case RED:
    case GREEN:
    case BLUE: hft_make_field_rgb(pixel_buffer); break;
    default: g_print("invalid color mode\n");
    }

  g_free(pixel_buffer);
}  

inline void hft_make_field_value(guchar* pixel_buffer)
{
  guint row, col;

  switch (hft_drawable->bpp)
    {
    case 4:
    case 3:
      for(row = 0; row < hft_current.rowspan; row++)
	{
	  for (col = 0; col< hft_current.colspan; col++)
	    { 
	     hft_current.field[ (hft_current.colspan * row) + col] = 	
		(pixel_buffer
		 [(row * hft_current.row_sample_rate * hft_drawable->width 
		   * hft_drawable->bpp) 
		 + (col * hft_current.col_sample_rate 
		    * hft_drawable->bpp)] 
		 + pixel_buffer
		 [(row * hft_current.row_sample_rate * hft_drawable->width 
		   * hft_drawable->bpp) 
		 + (col * hft_current.col_sample_rate 
		    * hft_drawable->bpp) + 1] 
		 + pixel_buffer
		 [(row * hft_current.row_sample_rate * hft_drawable->width 
		   * hft_drawable->bpp) 
		 + (col * hft_current.col_sample_rate 
		    * hft_drawable->bpp) + 2]) / hft_drawable->bpp;
	    }
	}
      break;

    case 2:
      for(row = 0; row < hft_current.rowspan; row++)
	{
	  for (col = 0; col< hft_current.colspan; col++)
	    { 
	      hft_current.field[ (hft_current.colspan * row) + col] = 
		pixel_buffer
		[(row * hft_current.row_sample_rate * hft_drawable->width) 
		+ (col * hft_current.col_sample_rate * hft_drawable->bpp)];
	    }
	}
      break;     

    case 1:
      for(row = 0; row < hft_current.rowspan; row++)
	{
	  for (col = 0; col< hft_current.colspan; col++)
	    { 
	      hft_current.field[ (hft_current.colspan * row) + col] = 
		pixel_buffer
		[(row * hft_current.row_sample_rate * hft_drawable->width)
		+ (col * hft_current.col_sample_rate)];
	    }
	}
      break;

    default: 
      g_print ("invalid bit depth\n");
    }
}

inline void hft_make_field_alpha(guchar* pixel_buffer)
{
  guint row, col;

  switch (hft_drawable->bpp)
    {
    case 4:
      for(row = 0; row < hft_current.rowspan; row++)
	{
	  for (col = 0; col< hft_current.colspan; col++)
	    { 
	      hft_current.field[ (hft_current.colspan * row) + col] = 
		pixel_buffer
		[(row * hft_current.row_sample_rate * hft_drawable->width * 4) 
		+ (col * hft_current.col_sample_rate * 4) + 3];
	    }
	}
      break;     
    case 2:
      for(row = 0; row < hft_current.rowspan; row++)
	{
	  for (col = 0; col< hft_current.colspan; col++)
	    { 
	      hft_current.field[ (hft_current.colspan * row) + col] = 
		pixel_buffer
		[(row * hft_current.row_sample_rate * hft_drawable->width * 2) 
		+ (col * hft_current.col_sample_rate * 2) + 1];
	    }
	}
      break;     

    default: 
      g_print ("invalid bit depth\n");
    }
}


inline void hft_make_field_rgb(guchar* pixel_buffer)
{
  guint row, col;
  gint color;

  switch (hft_current.displacement_mode)
    {
    case RED: color = 0; break;
    case GREEN: color = 1; break;
    default: color = 2; 
    }

  for(row = 0; row < hft_current.rowspan; row++)
    {
      for (col = 0; col< hft_current.colspan; col++)
	{ 
	  hft_current.field[ (hft_current.colspan * row) + col] = 
	    pixel_buffer
	    [ (row * hft_current.row_sample_rate * hft_drawable->width 
	       * hft_drawable->bpp) 
	    + (col * hft_current.col_sample_rate * hft_drawable->bpp) 
	    + color
	    ];
	}
    }
}

void hft_get_gradient()
{
  gdouble *double_gradient;
  gint i,j;

  if (hft_current.gradient == NULL)
    {
      hft_current.gradient = g_malloc( sizeof(guchar) * 4 * 256 );
      if (hft_current.gradient == NULL)
	{
	  g_print("could not allocate byte gradient in hft_get_gradient\n");
	  gtk_main_quit();
	}
    }

  double_gradient = gimp_gradients_sample_uniform(256);
  if (double_gradient == NULL)
    {
      g_print("could not allocate double gradient in hft_get_gradient\n");
      gtk_main_quit();
    } 
  
  for(i=0;i<256;i++)
    for (j=0;j<4;j++)
      hft_current.gradient[(i*4)+j] 
	= (guchar)(double_gradient[(i*4)+j] * 255);     
  g_free(double_gradient);
}

void hft_draw()
{
  GLfloat transform[4][4];

  if (gtk_gl_area_begingl(GTK_GL_AREA(hft_glarea))) 
    {
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
      
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(hft_current.zoom,1.0,0.3,    
		     hft_drawable->height * 3);  
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();       
      glTranslatef(-eye[X], -eye[Y], -eye[Z]);
      build_rotmatrix(transform,hft_current.quat);
      glMultMatrixf(&transform[0][0]);
      if (transform[2][2] < 0)
	if (transform[0][2] < 0)
	  quadrant = 2;
	else
	  quadrant = 3;
      else if (transform[0][2] < 0)
	quadrant = 1;
      else
	  quadrant = 0;
      switch (hft_current.display_mode) 
	{
	case FOREGROUND_POINT: 
	  hft_draw_foreground_point(); break;
	case GRADIENT_POINT:
	  hft_draw_gradient_point(); break; 
	case IMAGE_POINT:
	  hft_draw_image_point(); break;
	case IMAGE_LINE:
	  hft_draw_image_line(); break;
	case IMAGE_FILL:
	  hft_draw_image_fill(); break;
	case FOREGROUND_LINE:
	   hft_draw_foreground_line(); break;
	case GRADIENT_LINE:
	  hft_draw_gradient_line(); break;
	case GRADIENT_FILL:
	  hft_draw_gradient_fill(); break;	  
	default:
	  g_print("switch(hft_current.display_color) failed in hft_draw() ");
	  g_print("in hft_view.c\nplease email pixel@slimey.com\n");
	  gtk_main_quit();
	}
      gtk_gl_area_endgl(GTK_GL_AREA(hft_glarea));
    }
  gtk_gl_area_swapbuffers(GTK_GL_AREA(hft_glarea));
}  

void hft_draw_foreground_point() 
{
  gint row,col;
  guchar red, green, blue;
  gimp_palette_get_foreground (&red, &green, &blue);
  glColor3ub(red,green,blue);      
  glBegin(GL_POINTS);
  for (row = 0;row < hft_current.rowspan; row++)
    for (col = 0;col < hft_current.colspan; col++)
      place_vertex(row,col);
  glEnd();
}

void hft_draw_image_point()
{
  gint row,col;
  glBegin(GL_POINTS); 
      for (row = 0;row < hft_current.rowspan; row++)  
	{
	  for (col = 0;col < hft_current.colspan; col++) 
	    {
	      if ( (hft_current.field[(hft_current.colspan*row)+col]) 
		   > GRAY_THRESHOLD )
		{
		  glColor3ub(
			     hft_current.field[(hft_current.colspan* row)+col],
			     hft_current.field[(hft_current.colspan* row)+col],
			     hft_current.field[(hft_current.colspan* row)+col]
			     );
		  place_vertex(row,col);
		}
	    }
	}
  glEnd();
}

void hft_draw_gradient_point() 
{
  gint row,col;
#ifdef DEBUG
  g_print ("quadrant %d\n",quadrant);
#endif  
  glBegin(GL_POINTS);
  switch (quadrant)
    {
    case 0:
      for (row = 0;row < hft_current.rowspan; row++)  
	  for (col = 0;col < hft_current.colspan; col++) 	    
	      gradient_point(row,col);
      break;
    case 1:
      for (row = 0;row < hft_current.rowspan; row++)
	  for (col = hft_current.colspan - 1; col >= 0; col--) 
	    gradient_point(row,col);
      break;
    case 2:
      for (row = hft_current.rowspan - 1; row >= 0; row--)
	  for (col = hft_current.colspan - 1; col >= 0; col--) 
	    gradient_point(row,col);
      break;
    case 3:
      for (row = hft_current.rowspan - 1; row >= 0; row--)
	  for (col = 0;col < hft_current.colspan; col++) 
	    gradient_point(row,col);
      break;
    }
  glEnd();
}

void hft_draw_foreground_line() 
{
  gint row,col;
  guchar red, green, blue;
  gimp_palette_get_foreground (&red, &green, &blue);
  glColor3ub(red,green,blue);      
  for (row = 0;row < hft_current.rowspan-1; row++)
    {
      glBegin(GL_TRIANGLE_STRIP);  
      for (col = 0;col < hft_current.colspan-1; col++)
	{
	  place_vertex(row+1,col); 
	  place_vertex(row,col);
	}
      glEnd();
    }
}

void hft_draw_image_line() 
{
  gint row,col;
  for (row = 0;row < hft_current.rowspan-1; row++)
    {
      glBegin(GL_TRIANGLE_STRIP);  
      for (col = 0;col < hft_current.colspan-1; col++)
	{
	  glColor3ub(
		     hft_current.field[(hft_current.colspan* row)+col],
		     hft_current.field[(hft_current.colspan* row)+col],
		     hft_current.field[(hft_current.colspan* row)+col]
		     );
	  place_vertex(row+1,col); 
	  place_vertex(row,col);
	}
      glEnd();
    }
}

void hft_draw_image_fill() 
{
  gint row,col;
  for (row = 0;row < hft_current.rowspan-1; row++)
    {
      glBegin(GL_TRIANGLE_STRIP);  
      for (col = 0;col < hft_current.colspan-1; col++)
	{
	  glColor3ub(
		     hft_current.field[(hft_current.colspan* row+1)+col],
		     hft_current.field[(hft_current.colspan* row+1)+col],
		     hft_current.field[(hft_current.colspan* row+1)+col]
		     );
	  place_vertex(row+1,col); 
	  glColor3ub(
		     hft_current.field[(hft_current.colspan* row)+col],
		     hft_current.field[(hft_current.colspan* row)+col],
		     hft_current.field[(hft_current.colspan* row)+col]
		     );
	  place_vertex(row,col);
	}
      glEnd();
    }
}

void hft_draw_gradient_fill() 
{
  gint row,col;
#ifdef DEBUG
  g_print ("quadrant %d\n",quadrant);
#endif  
  switch (quadrant)
    {
    case 0:
      for (row = 0;row < hft_current.rowspan - 1; row++)  
	{
	  glBegin(GL_TRIANGLE_STRIP);
	  for (col = 0;col < hft_current.colspan-1; col++) 
	    {
	      gradient_point(row+1,col);
	      gradient_point(row,col);
	    }
	  glEnd();
	}    
    case 1:
      for (row = 0;row < hft_current.rowspan - 1; row++)  
	{
	  glBegin(GL_TRIANGLE_STRIP);
	  for (col = hft_current.colspan-1; col >=0; col--) 
	    {
	      gradient_point(row+1,col);
	      gradient_point(row,col);
	    }
	  glEnd();
	}    
    case 2:
      for (row = hft_current.rowspan - 1; row >= 1;row--)  
	{
	  glBegin(GL_TRIANGLE_STRIP);
	   for (col = hft_current.colspan-1; col >=0; col--)
	    {
	      gradient_point(row-1,col);
	      gradient_point(row,col);
	    }
	  glEnd();
	}    
    case 3:
      for (row = hft_current.rowspan - 1; row >= 1;row--)   
	{
	  glBegin(GL_TRIANGLE_STRIP);
	  for (col = 0;col < hft_current.colspan-1; col++) 
	    {
	      gradient_point(row-1,col);
	      gradient_point(row,col);
	    }
	  glEnd();
	}
    }
}

void hft_draw_gradient_line() 
{
  gint row,col;
  for (row = 0;row < hft_current.rowspan-1; row++)  
    {
      glBegin(GL_TRIANGLE_STRIP);
      for (col = 0;col < hft_current.colspan-1; col++) 
	{
	  gradient_point(row+1,col);
	  gradient_point(row,col);
	}
      glEnd();
    }
}

inline void gradient_point(int row, int col)
{
  glColor3ub( hft_current.gradient 
	      [ hft_current.field
	      [ ( hft_current.colspan * row ) + col ] * 4 ],
	      
	      hft_current.gradient
	      [ hft_current.field
	      [ ( hft_current.colspan * row ) + col ] * 4 + 1],
	      
	      hft_current.gradient
	      [ hft_current.field
	      [ ( hft_current.colspan * row ) + col ] * 4 + 2]);
  
   place_vertex(row,col);
}

inline void place_vertex(int row, int col)
{
  glVertex3f( (col * hft_current.col_sample_rate)
	      - (GLfloat)(hft_drawable->width * 0.5),
	      
	      hft_current.field 
	      [(hft_current.colspan * row)+ col] 
	      * hft_drawable->width/256
	      * hft_current.displacement_scale
	      - (GLfloat)(hft_drawable->width * 0.5) 
	      * hft_current.displacement_scale,  
	      
	      (row * hft_current.row_sample_rate)
	      - (GLfloat)(hft_drawable->height * 0.5) );
}

void motion_notify(GtkWidget *widget, GdkEventMotion *event)
{
  int x, y;
  GdkRectangle area;
  GdkModifierType state;
  float spin_quat[4];

  if (event->is_hint) 
    {
      gdk_window_get_pointer(event->window, &x, &y, &state);
    } 
  else 
    { 
      x = event->x;
      y = event->y;
      state = event->state;
    }
  
  area.x = 0;
  area.y = 0;
  area.width  = widget->allocation.width;
  area.height = widget->allocation.height;

  /* pan */
  if (state & GDK_BUTTON1_MASK) {
    eye[X] -= x - hft_current.x; 
    eye[Y] += y - hft_current.y; 
    gtk_widget_draw(widget, &area); 
  }

  /* rotate */
  if (state & GDK_BUTTON2_MASK) {
     trackball(spin_quat,
	      (2.0*hft_current.x -       area.width) / area.width,
	      (     area.height - 2.0*hft_current.y) / area.height,
	      (           2.0*x -       area.width) / area.width,
	      (     area.height -            2.0*y) / area.height);
    add_quats(spin_quat, hft_current.quat, hft_current.quat);
    gtk_widget_draw(widget, &area);
  }
 
  /* zoom */
  if (state & GDK_BUTTON3_MASK) {
    hft_current.zoom += ((y - hft_current.y) / area.height) * 40;
    if (hft_current.zoom < 5) hft_current.zoom = 5;
    if (hft_current.zoom > 120) hft_current.zoom = 120;
    gtk_widget_draw(widget, &area); 
  }

  hft_current.x = x;
  hft_current.y = y;
}

static void button_press() 
{     
#ifdef DEBUG
  g_print("button_press\n");
#endif
  set_auto_update(FALSE);
}

static void button_release()
{
#ifdef DEBUG
  g_print("button_release\n");
#endif
  set_auto_update(TRUE);
}    

static gint expose_callback(GtkWidget *widget, GdkEventExpose *event)
{
  if (event->count > 0) return TRUE;
  hft_draw(NULL);
  return TRUE;
}

static gint reshape(GtkWidget *widget, GdkEventConfigure *event)
{
  if (gtk_gl_area_begingl(GTK_GL_AREA(widget))) 
    {    
      glViewport(0,0, widget->allocation.width, widget->allocation.height);
      gtk_gl_area_endgl(GTK_GL_AREA(widget));
    }
  return TRUE;
}


